%include "./utils/bootstrap.asm"
%include "./utils/protected_mode.asm"
%include "./utils/pci.asm"

BLACK         equ 0x0
BLUE          equ 0x1
GREEN         equ 0x2
CYAN          equ 0x3
RED           equ 0x4
MAGENTA       equ 0x5
BROWN         equ 0x6
LIGHT_GREY    equ 0x7
DARK_GREY     equ 0x8
LIGHT_BLUE    equ 0x9
LIGHT_GREEN   equ 0xA
LIGHT_CYAN    equ 0xB
LIGHT_RED     equ 0xC
LIGHT_MAGENTA equ 0xD
LIGHT_BROWN   equ 0xE
WHITE         equ 0xF

VGA_ROW_SIZE equ 80
COLOR equ BLUE << 4 | GREEN
; jmp read_trb
; Check if BAR_1 is not empty (this code is not ready for that case)
call find_xhci
mov al, PCI_BAR_1_OFFSET
read_pci_register
cmp eax, 0x00
je next
mov esi, error_str
mov edi, VGA_ROW_SIZE*0
mov dh, RED << 4 | WHITE
call print_str_PM
jmp end

; Get capability register's address
next:
call find_xhci
mov al, PCI_BAR_0_OFFSET
read_pci_register
and eax, 0xFF_FF_FF_F0

; Note that we are not reading PCI_BAR_1_OFFSET
; because this example is set up to work only
; with 32 bit addresses

; Get operational register's address
mov ebx, eax        ; BASE_ADDRESS (== CAPLENGTH)
mov bl, [es:ebx]    ; [CAPLENGTH]
and ebx, 0x00_00_00_FF
add ebx, eax        ; BASE_ADDRESS + [CAPLENGTH] = OPERATIONAL



; STOP THE XHCI

; Read the contents of USBCMD register
mov ecx, [es:ebx]
; Set the R/S bit to '0'
and ecx, 0b1111_1111_1111_1111__1111_1111_1111_1110
; Write the USBCMD register
mov [es:ebx], ecx



; CHECK THE XHCI IS STOPPED

check_HCH_bit:
; Read the USBSTS register
mov ecx, [es:ebx + 0x04]
; Check the HCH bit
and ecx, 0b0000_0000_0000_0000__0000_0000_0000_0001
jz check_HCH_bit



; RESET THE XHCI

; Read the contents of USBCMD register
mov ecx, [es:ebx]
; Set the HCRST bit to 1
or ecx, 0b0000_0000_0000_0000__0000_0000_0000_0010
; Write back the USBCMD register
mov [es:ebx], ecx



; CHECK THE RESET IS COMPLETED

check_CNR_bit:
; Read the USBSTS register
mov ecx, [es:ebx + 0x04]
; Check the CNR bit
and ecx, 0b0000_0000_0000_0000__0000_1000_0000_0000
jnz check_CNR_bit

check_HCRST_bit:
; Read the contents of USBCMD register
mov ecx, [es:ebx]
; Check the HCRST bit is zero
and ecx, 0b0000_0000_0000_0000__0000_0000_0000_0010
jnz check_HCRST_bit

check_CNR_bit_2:
; Read the USBSTS register
mov ecx, [es:ebx + 0x04]
; Check the CNR bit
and ecx, 0b0000_0000_0000_0000__0000_1000_0000_0000
jnz check_CNR_bit_2



; CHECK 64 BITS ALIGMENT

mov ecx, NOOP_TRB
and ecx, 0b0000_0000_0000_0000__0000_0000_0011_1111
jz next2
; Print ERROR
mov esi, error_str
mov edi, VGA_ROW_SIZE*0
mov dh, RED << 4 | WHITE
call print_str_PM
jmp end
next2:



; DEFINE COMMAND RING DEQUEUE POINTER

mov ecx, [es:ebx + 0x18] ; CRCR
;                                                                     Command Ring Pointer                  RsvdP   CRR     CA  CS  RSC
and ecx,                                                            0b0000_0000_0000_0000_0000_0000_00______11______1_______0___0___1
or ecx,     (0x00_00_00_00__FF_FF_FF_FF & ((NOOP_TRB-$$)+0x7C00)) | 0b0000_0000_0000_0000_0000_0000_00______00______0_______0___0___1
mov         [es:ebx + (0x18+4*0)], ecx
mov dword   [es:ebx + (0x18+4*1)], ((NOOP_TRB-$$)+0x7C00) >> 32



; PROGRAM THE ERSTSZ REGISTER

mov ebx, eax                    ; BASE_ADDRESS
mov ebx, [es:ebx + 0x18]        ; [RTSOFF]
add ebx, eax                    ; BASE_ADDRESS + [RTSOFF] = RUNTIME
mov word [es:ebx + 0x28], 0x01  ; RUNTIME + 0x28 = ERSTSZ



; CHECK 16 BITS ALIGMENT

mov ecx, event_ring
and ecx, 0b0000_0000_0000_0000__0000_0000_0000_1111
jz next3
; Print ERROR
mov esi, error_str
mov edi, VGA_ROW_SIZE*0
mov dh, RED << 4 | WHITE
call print_str_PM
jmp end
next3:


; PROGRAM THE ERDP REGISTER

;                                                      EHB  DESI
mov word [es:ebx + 0x38], ((event_ring-$$)+0x7C00) | 0b0____000



; CHECK 64 BITS ALIGMENT

mov ecx, event_ring_segment_table
and ecx, 0b0000_0000_0000_0000__0000_0000_0011_1111
jz next4
; Print ERROR
mov esi, error_str
mov edi, VGA_ROW_SIZE*0
mov dh, RED << 4 | WHITE
call print_str_PM
jmp end
next4:



; PROGRAM THE ERSTBA

mov ecx, [es:ebx + 0x30] ; ERSTBA
and ecx, 0b0000_0000_0000_0000__0000_0000_0011_1111
or ecx, 0x00_00_00_00__FF_FF_FF_FF & ((event_ring_segment_table-$$)+0x7C00)
mov [es:ebx + (0x30 + 4*0)], ecx
mov dword [es:ebx + (0x30 + 4*1)], ((event_ring_segment_table-$$)+0x7C00) >> 32



; START THE XHCI

; Read the contents of USBCMD register
mov ecx, [es:ebx]
; Set the R/S bit to '1'
or ecx, 0b0000_0000_0000_0000__0000_0000_0000_0001
; Write the USBCMD register
mov [es:ebx], ecx



; CHECK CONTROLLER NOT READY (CNR) BIT

check_CNR_bit_3:
; Read the USBSTS register
mov ecx, [es:ebx + 0x04]
; Check the CNR bit
and ecx, 0b0000_0000_0000_0000__0000_1000_0000_0000
jnz check_CNR_bit_3



; RING THE DOORBELL

mov ebx, eax                    ; BASE_ADDRESS
mov ebx, [es:ebx + 0x14]        ; [DBOFF]
add ebx, eax                    ; BASE_ADDRESS + [RTSOFF] = DOORBELL
mov byte [es:ebx], 0x00


; READ EVENT TRB

read_trb:
mov edx, 0

mov cx, 32
read_trb.A:
dec cx
mov eax, [es:event_ring + 4*3]
shr eax, cl
and eax, 1
add al, '0'
mov ah, COLOR
mov [es:edx*2 + 0xB8000], ax
inc edx
cmp cx, 0
jne read_trb.A

; 32 bits separator
mov word [es:edx*2 + 0xB8000], COLOR << 8 | '_'
inc edx

mov cx, 32
read_trb.B:
dec cx
mov eax, [es:event_ring + 4*2]
shr eax, cl
and eax, 1
add al, '0'
mov ah, COLOR
mov [es:edx*2 + 0xB8000], ax
inc edx
cmp cx, 0
jne read_trb.B

; 32 bits separator
mov word [es:edx*2 + 0xB8000], COLOR << 8 | '_'
inc edx

mov cx, 32
read_trb.C:
dec cx
mov eax, [es:event_ring + 4*1]
shr eax, cl
and eax, 1
add al, '0'
mov ah, COLOR
mov [es:edx*2 + 0xB8000], ax
inc edx
cmp cx, 0
jne read_trb.C

; 32 bits separator
mov word [es:edx*2 + 0xB8000], COLOR << 8 | '_'
inc edx

mov cx, 32
read_trb.D:
dec cx
mov eax, [es:event_ring + 4*0]
shr eax, cl
and eax, 1
add al, '0'
mov ah, COLOR
mov [es:edx*2 + 0xB8000], ax
inc edx
cmp cx, 0
jne read_trb.D
jmp read_trb

end:
    cli
    hlt
    jmp end

%include "./utils/print.asm"
%include "./utils/find_xhci.asm"

success_str db "SUCCESS!!", 0
error_str db "ERROR", 0

times 64 - ((($-$$)+0x7C00) % 64) db 0

NOOP_TRB:
dd 0x00
dd 0x00
dd 0x00
; TRB Type = 23 = 0b10111
;    TRB Type   RsvdZ   IOC CH  RsvdZ   ENT C
dw 0b010111_____0000____1___0___00______0___0
dw 0x00

; This TRB has the Cycle bit (C) set to zero
; therefore it must not be executed by xHCI
dq 0x00
dq 0x00

times 64 - ((($ - $$) + 0x7C00) % 64) db 0

event_ring_segment_table:
dq event_ring
dw 64
dw 0x00

times 64 - ((($ - $$) + 0x7C00) % 64) db 0

event_ring:
times 0x1000 db 0

; The following is there to check that the
; correct number of sectors are loaded
times 512*(NUMBER_OF_SECTORS+1)-($-$$) db 0